"""Run a OpenFOAM command"""

import sys
import string
from os import path

from FoamThread import FoamThread
from PyFoam.Infrastructure.FoamServer import FoamServer
from PyFoam.Infrastructure.Logging import foamLogger
from PyFoam.RunDictionary.SolutionDirectory import SolutionDirectory
from PyFoam.Execution.ParallelExecution import LAMMachine
from PyFoam.RunDictionary.ParameterFile import ParameterFile

class BasicRunner(object):
    """Base class for the running of commands

    When the command is run the output is copied to a LogFile and
    (optionally) standard-out

    The argument list assumes for the first three elements the
    OpenFOAM-convention:

    <cmd> <dir> <case>

    The directory name for outputs is therefor created from <dir> and
    <case>
    
    Provides some handle-methods that are to be overloaded for
    additional functionality"""
    
    def __init__(self,argv=None,silent=False,logname="PyFoam",lam=None,server=False):
        """@param argv: list with the tokens that are the command line
        if not set the standard command line is used
        @param silent: if True no output is sent to stdout
        @param logname: name of the logfile
        @param lam: Information about a parallel run
        @param server: Whether or not to start the network-server
        @type lam: PyFoam.Execution.ParallelExecution.LAMMachine"""
        
        if argv==None:
            self.argv=sys.argv[1:]
        else:
            self.argv=argv
        self.dir=path.join(self.argv[1],self.argv[2])
        if self.argv[2][-1]==path.sep:
            self.argv[2]=self.argv[2][:-1]
            
        self.silent=silent
        self.lam=lam
        self.origArgv=self.argv
        
        if self.lam!=None:
            self.argv=lam.buildMPIrun(self.argv)
        self.cmd=string.join(self.argv," ")
        foamLogger().info("Starting: "+self.cmd+" in "+path.abspath(path.curdir))
        self.logFile=path.join(self.dir,logname+".logfile")
        
        self.fatalError=False
        self.warnings=0
        self.started=False

        self.run=FoamThread(self.cmd)

        self.server=None
        if server:
            self.server=FoamServer(run=self.run,master=self)
            self.server.setDaemon(True)
            self.server.start()

        self.createTime=None
        self.nowTime=None

        self.stopMe=False
        
    def start(self):
        """starts the command and stays with it till the end"""
        
        self.started=True
        fh=open(self.logFile,"w")

        self.startHandle()

        check=BasicRunnerCheck()
        
        self.run.start()

        while self.run.check():
            try:
                self.run.read()
                if not self.run.check():
                    break

                line=self.run.getLine()

                tmp=check.getTime(line)
                if tmp!=None:
                    self.nowTime=tmp
                    if self.createTime==None:
                        # necessary because interFoam reports no creation time
                        self.createTime=tmp
                        
                tmp=check.getCreateTime(line)
                if tmp!=None:
                    self.createTime=tmp
                    
                if not self.silent:
                    print line
                    
                if line.find("FOAM FATAL ERROR")>=0 or line.find("FOAM FATAL IO ERROR")>=0:
                    self.fatalError=True

                if self.fatalError and line!="":
                    foamLogger().error(line)

                if line.find("FOAM Warning")>=0:
                    self.warnings+=1

                if self.server!=None:
                    self.server._insertLine(line)
                
                self.lineHandle(line)

                fh.write(line+"\n")
                fh.flush()
            except KeyboardInterrupt,e:
                foamLogger().warning("Keyboard Interrupt")
                self.run.interrupt()
                
        self.stopHandle()
        
        fh.close()
        
        if self.server!=None:
            self.server.deregister()
            
        foamLogger().info("Finished")

    def runOK(self):
        """checks whether the run was successful"""
        if self.started:
            return not self.fatalError
        else:
            return False
        
    def startHandle(self):
        """to be called before the program is started"""
        pass

    def stopGracefully(self):
        """Tells the runner to stop at the next convenient time"""
        if not self.stopMe:
            self.stopMe=True
            self.controlDict=ParameterFile(path.join(self.dir,"system","controlDict"),backup=True)
            dt=self.controlDict.readParameter("deltaT")
            self.controlDict.replaceParameter("stopAt","nextWrite")
            self.controlDict.replaceParameter("writeInterval",dt)
            print "Stopping run at next write"
        
    def stopHandle(self):
        """called after the program has stopped"""
        if self.stopMe:
            self.controlDict.restore()
    
    def lineHandle(self,line):
        """called every time a new line is read"""
        pass
    
    def logName(self):
        """Get the name of the logfiles"""
        return self.logFile

    def getSolutionDirectory(self,archive=None):
        """@return: The directory of the case
        @rtype: PyFoam.RunDictionary.SolutionDirectory
        @param archive: Name of the directory for archiving results"""

        return SolutionDirectory(self.dir,archive=archive)

import re

class BasicRunnerCheck(object):
    """A small class that does primitve checking for BasicRunner
    Duplicates other efforts, but ...."""

    floatRegExp="[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?"

    def __init__(self):
        self.timeExpr=re.compile("^Time = (%f%)$".replace("%f%",self.floatRegExp))
        self.createExpr=re.compile("^Create mesh for time = (%f%)$".replace("%f%",self.floatRegExp))
        
    def getTime(self,line):
        """Does this line contain time information?"""
        m=self.timeExpr.match(line)
        if m:
            return float(m.group(1))
        else:
            return None
        
    def getCreateTime(self,line):
        """Does this line contain mesh time information?"""
        m=self.createExpr.match(line)
        if m:
            return float(m.group(1))
        else:
            return None
        
