"""Analyze OpenFOAM logs"""

from LogLineAnalyzer import LogLineAnalyzer
from TimeLineAnalyzer import TimeLineAnalyzer
from PyFoam.Basics.LineReader import LineReader
from PyFoam.Error import error

import sys

class FoamLogAnalyzer(object):
    """Base class for all analyzers

    Administrates and calls a number of LogLineAnlayzers for each
    line"""
    
    def __init__(self,progress=False):
        """
        @param progress: Print time progress on console?
        """
        self.analyzers={}
        self.time=""
        self.oDir=""
        self.line=LineReader()
        self.timeListeners=[]
        
        self.addAnalyzer("Time",TimeLineAnalyzer(progress=progress))

    def tearDown(self):
        """Remove reference to self in children (hoping to remove
        circular dependencies)"""
        
        for a in self.analyzers.values():
            a.setParent(None)
            
    def setTime(self,time):
        """Sets the time and alert all the LineAnalyzers that the time has changed
        @param time: the new value of the time
        """
        if time!=self.time:
            self.time=time
            for listener in self.timeListeners:
                listener.timeChanged()
            for nm in self.analyzers:
                self.analyzers[nm].timeChanged()

    def addTimeListener(self,listener):
        """@param listener: An object that is notified when the time changes. Has to
        implement a timeChanged method"""
        if not 'timeChanged' in dir(listener):
            error("Error. Object has no timeChanged-method:"+str(listener))
        else:
            self.timeListeners.append(listener)

    def listAnalyzers(self):
        """@returns: A list with the names of the Analyzers"""
        return self.analyzers.keys()
    
    def getAnalyzer(self,name):
        """Get the LogLineAnalyzer name"""
        if self.analyzers.has_key(name):
            return self.analyzers[name]
        else:
            return None
        
    def addAnalyzer(self,name,obj):
        """Adds an analyzer

        obj - A LogLineAnalyzer
        name - the name of the analyzer"""
        
        obj.setParent(self)
        self.analyzers[name]=obj

    def analyzeLine(self,line):
        """Calls all the anlyzers for a line"""
        for nm in self.analyzers:
            self.analyzers[nm].doAnalysis(line)

    def analyze(self,fh):
        """Analyzes a file (one line at a time)

        fh - handle of the file"""
        while(self.line.read(fh)):
            self.analyzeLine(self.line.line)

    def goOn(self):
        """Checks with all the analyzers

        If one analyzer returns False it returns False"""
        result=True

        for nm in self.analyzers:
            #            print nm,self.analyzers[nm].goOn()
            result=result and self.analyzers[nm].goOn()
        
        return result
    
    def getTime(self):
        """Gets the current time"""
        return str(self.time)

    def setDirectory(self,d):
        """Sets the output directory for all the analyzers"""
        self.oDir=d
        for nm in self.analyzers:
            self.analyzers[nm].setDirectory(self.oDir)

    def getDirectory(self):
        """Gets the output directory"""
        return self.oDir
    

    
